# -*- coding:utf-8 -*-
#---------------------------------------------------------------------
#
# Qualitative Spatial Reasoning for routing
#
# Copyright (C) 2011  Germán Carrillo (http://geotux.tuxfamily.org)
#
#---------------------------------------------------------------------
#
# licensed under the terms of GNU GPL 2
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this software. If not, see <http://www.gnu.org/licenses/>.
#
#---------------------------------------------------------------------
import os
from PyQt4.QtGui import *

class Algorithm():
    def __init__(self, graph):
        self.graph = graph # Graph data structure with the network data
        self.sparq_dir = "/docs/downloads/sparq/" # SparQ installation dir
        self.route = {} # Dict selected dipole : Alternatives for this dipole
        self.route_order = [] # To store the ordered route (dict doesn't allow it)
        self.next_dipoles = {} # Set of next dipoles to choose from, dipole: consistent
        self.bad_dipoles_list = [] # Dipoles that ONLY lead to already visited dipoles or to nowhere
        self.msg = MessagesManager('normal') # Options: normal | debug


    def getMessageToPrint(self):
        return self.msg.msgToPrint

    def run(self, starting_dipole, target_dipole):
        sd= starting_dipole #"10-12" # Starting dipole 
        td= target_dipole # "135-154" # Target dipole 
        cd = sd # Current dipole
        cn = '' # Current node
        self.next_dipoles[sd] = True
        step = 0
        while True: 
            step += 1
            self.msg.imprimir( "************ STEP " + str( step ) + " **************" ) 

            # Select the "next" current dipole
            next_dipole_found = False
            if len(self.next_dipoles) > 0: 

                # Check if any of the next_dipoles is connected (share a node) to td
                for k,v in self.next_dipoles.iteritems():
                    td_n1, td_n2 = td.split("-")[0], td.split("-")[1]
                    k_n1, k_n2 = k.split("-")[0], k.split("-")[1]
                    
                    # td is connected to k, we are done!
                    if td_n1 == k_n1 or td_n1 == k_n2 or td_n2 == k_n1 or td_n2 == k_n2:
                        self.route_order.append(k) 
                        self.route_order.append(td)
                        self.msg.imprimir( "I: Candidate " + k + " connected with target dipole!" )
                        self.msg.imprimir( "\n*** Eureka! The route is:" + str(  self.route_order ) + "***" )
                        return self.route_order

                # Pick the next dipole and remove it as alternative dipole
                for k,v in self.next_dipoles.iteritems():
                    if self.next_dipoles[k] == True: # Consistent
                        if k not in self.route_order and k not in self.bad_dipoles_list: # Don't pick an already chosen dipole
                            cd = k
                            self.next_dipoles.pop(k)
                            self.route[cd] = self.next_dipoles.keys() # Take the alternatives 
                            next_dipole_found = True
                            break
                else: # Ok, just pick a non consistent one
                    for k,v in self.next_dipoles.iteritems():
                        if self.next_dipoles[k] == False: # Not consistent
                            if k not in self.route_order and k not in self.bad_dipoles_list: # Don't pick an already chosen dipole
                                cd = k   
                                self.next_dipoles.pop(k)
                                self.route[cd] = self.next_dipoles.keys() # Take the alternatives
                                next_dipole_found = True
                                break

            if not next_dipole_found: # Whether no candidates or all already chosen, go back in the route   
                alternative_found = False

                while not alternative_found and len(self.route_order) > 0:         
                    cd = self.route_order.pop() # Remove the last dipole visited from the route
                    next_alternatives = self.route.pop(cd) # and from the alternatives dict

                    if len(next_alternatives) > 0:
                        for a in next_alternatives:
                            if a not in self.route_order and a not in self.bad_dipoles_list: # Don't pick an already chosen dipole
                                cd = a
                                next_alternatives.remove(a)
                                self.route[cd] = next_alternatives
                                self.msg.imprimir( "I: No options found, going back to dipole... " + cd )   
                                alternative_found = True

                if not alternative_found: 
                    self.msg.imprimir( "Sorry, we couldn't find a way. No. of steps: " + str(step) )
                    return None 

            self.route_order.append(cd)
            self.msg.imprimir( "Current dipole:" + str( cd ) )

            # Get a node from the current dipole
            # Keep both nodes in the case we have to switch them
            temp_node = self.graph.getEdge(cd).n2.id 
            if temp_node == cn: # Don't get the same node again
                cn = self.graph.getEdge(cd).n1.id 
            else:
                cn = temp_node
                temp_node = self.graph.getEdge(cd).n1.id

            edges = self.graph.getEdges(cn) # Get candidates from the current node

            # We got the starting dipole and the end node is not connected to anything 
            if len(edges) == 1 and len(self.route_order) == 1: 
                cn = temp_node # Switch the nodes from the starting dipole
                edges = self.graph.getEdges(cn) # Get candidates from this node

            self.msg.imprimir( "Current node:" + str( cn ) )
            r_cd_td = self.sparq_qualify(cd, td) # Relations cd-td
            #print "r_cd_td:", r_cd_td          

            for e in edges: 
                if td == e.id: # td is connected to cd, we are done!
                    self.route_order.append(td)
                    self.msg.imprimir( "\n*** Eureka! The route is:" + str(  self.route_order ) + "***" )
                    return self.route_order

            for e in edges: 
                if e.id in self.route_order or e.id in self.bad_dipoles_list: # Already chosen dipole?
                    edges.remove(e)            
            
            for e in edges:
                if cd == e.id: #Remove the current dipole from the list of candidates
                    edges.remove(e)

            for e in edges: 
                self.msg.imprimir( "Candidate: " + str( e.id ) )

            self.next_dipoles={}
            if len(edges) == 0:
                self.bad_dipoles_list.append(cd) # no alternatives found, avoid this dipole in the future
            elif len(edges) == 1:
                self.next_dipoles[edges[0].id] = True
                self.msg.imprimir( "I: No need to check consistency, only one option to be followed!" )
            else:
                r_c_cd = self.sparq_qualify_all(cd, edges) # Relations candidates - cd
                #print "r_c_cd:",r_c_cd
                for k,v in r_c_cd.iteritems():
                    cc = k.split("_")[1]

                    if v == True: # Workaround fot the dipoles-lying-in-the-same-line problem
                        self.next_dipoles[cc] = True
                    else: # Appropriate consistency check
                        if self.sparq_consistent(v,r_cd_td,cc,td): # Check consistency
                            self.next_dipoles[cc] = True
                        else:
                            self.next_dipoles[cc] = False # Not consistent
            self.msg.imprimir( "Consistency:" + str( self.next_dipoles ) )

        self.msg.imprimir( "Sorry, I couldn't find a way. No. of steps: " + str(step) )
        return None

    def sparq_qualify(self, d1, d2):
        """ Calls the SparQ module Qualify to describe a situation """
        d1_formatted = self.serialize_coord(d1)
        d2_formatted = self.serialize_coord(d2)
        
        # Check if the dipoles lie in the same line and avoid the command call
        d1_coord = d1_formatted[1:-1].split(" ")
        d2_coord = d2_formatted[1:-1].split(" ")
        d1p1x, d1p1y = float(d1_coord[1]), float(d1_coord[2])
        d1p2x, d1p2y = float(d1_coord[3]), float(d1_coord[4])
        d2p1x, d2p1y = float(d2_coord[1]), float(d2_coord[2])
        d2p2x, d2p2y = float(d2_coord[3]), float(d2_coord[4])
        m1,m2=0,0
        if d1p2x-d1p1x != 0: m1=(d1p2y-d1p1y)/(d1p2x-d1p1x) # Non-QSR (Just to handle a SparQ bug)
        if d2p2x-d2p1x != 0: m2=(d2p2y-d2p1y)/(d2p2x-d2p1x) # Non-QSR (Just to handle a SparQ bug)
        if m1==m2: 
            self.msg.imprimir( "Collinear dipoles handled!!!" )
            return True

        # dra-24 corresponds to Dipole Calculus
        cmd = self.sparq_dir + 'sparq qualify dra-24 all "(' + d1_formatted + d2_formatted + ')"'
        stream = os.popen(cmd)
        res = stream.read()
        return "(" + res.split(")")[0].split("(")[2] + ")"

    def serialize_coord(self, d):
        """ Helping function to get coordinates in SparQ syntaxis """
        e = self.graph.getEdge(d)
        x0=str(e.n1.x)
        y0=str(e.n1.y)
        x1=str(e.n2.x)
        y1=str(e.n2.y)
        
        return "(" + d + " " + x0 + " " + y0 + " " + x1 + " " + y1 + ")"

    def sparq_qualify_all(self, cd, dipoles):
        """ Calls the SparQ Qualify module for each pair candidate-cd """
        r_c_cd = {}        
        for d in dipoles:
            r_c_cd[cd+"_"+d.id] = self.sparq_qualify(cd,d.id)
        return r_c_cd

    def sparq_consistent(self, v, r_cd_td, cc, td):
        """ Checks scenario-consistency in SparQ from: 
            v: Relation current candidate - current dipole
            r_cd_td: Relation current dipole - target dipole
            cc: Current candidate
            td: Target dipole
        """
        constraint = "(ells errs lere rele)" # Our main contribution, the constraint!
        cmd = self.sparq_dir + 'sparq constraint-reasoning dra-24 scenario-consistency all "(' + v + r_cd_td + '(' + cc + ' ' + constraint + ' ' + td + '))"'
        stream = os.popen(cmd)
        res = stream.read()     
        if res != "Not consistent.\n\n":
            return True
        else:
            return False


class MessagesManager():

    def __init__(self, mode='normal'):
        self.mode = mode
        self.msgToPrint = []

    def imprimir(self, msg):
        if self.mode == 'debug':
            print msg
        else:
            self.msgToPrint.append( msg ) 

if __name__ == "__main__":
    import sys
    from PyQt4.QtGui import *
    from qgis.core import *
    from qgis.gui import *
    from graph import *

    qgis_path="/usr"
    app = QApplication( [] )
    QgsApplication.setPrefixPath( qgis_path, True )
    QgsApplication.initQgis()

    #layer = QgsVectorLayer("/docs/masteria/1st_semester/QSKRR/final_assignment/test/network.shp", "layer_name", "ogr")
    #layer = QgsVectorLayer("/docs/geodata/munster/streets/street_graph.shp", "layer_name", "ogr")
    layer = QgsVectorLayer("/docs/geodata/munster/osm/street_graph_osm.shp", "layer_name", "ogr")
    pr = layer.dataProvider()
    allAttrs = pr.attributeIndexes()
    layer.select(allAttrs)
    graph = Graph()
    for f in layer:
        line = f.geometry().asPolyline()
        n1 = graph.addNode(line[0].x(), line[0].y()) 
        n2 = graph.addNode(line[len(line)-1].x(), line[len(line)-1].y())
        graph.addEdge(n1, n2)
        graph.addRelFeatNodes(f.id(), n1, n2)

    algorithm = Algorithm( graph )
    args = sys.argv
    sd, td = "19-47", "59-68" # Custom dipole ids, overriden if you pass command-line parameters
    if len(args) == 3: 
        sd, td = args[1], args[2]
    route = algorithm.run( sd, td )

    if route: # Create memory layer and add it to QGIS
        for t in algorithm.getMessageToPrint():
            print t

    retval = app.exec_()

    # Salir
    QgsApplication.exitQgis()
    sys.exit( retval )

