# -*- coding:utf-8 -*-
#---------------------------------------------------------------------
# 
# Visor Geográfico
# An open source GIS application for IGAC's basic projects in municipalities.
# 
# Original sources Copyright (C) 2006 by Martin Dobias wonder.sk at gmail dot com
# Ported to Python by Agustín Codazzi Geographic Institute 2009
#
#---------------------------------------------------------------------
# 
# licensed under the terms of GNU GPL 2
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this software. If not, see <http://www.gnu.org/licenses/>.
# 
#---------------------------------------------------------------------

from PyQt4.QtCore import Qt, QRect, QPoint, SIGNAL
from PyQt4.QtGui import *

from qgis.core import *
from qgis.gui import *

class SeleccionarMapTool( QgsMapTool ): 
    def __init__( self, canvas, bStartingNode ):     
        QgsMapTool.__init__( self, canvas )
        self.canvas = canvas
        self.rubberBand = QgsRubberBand( self.canvas, False )
        self.activeLayer = None
        self.bStartingNode = bStartingNode

        self.cursor = QCursor( QPixmap( [  "16 16 3 1",
                                            "# c None",
                                            "a c #000000",
                                            ". c #ffffff",
                                            ".###############",
                                            "...#############",
                                            ".aa..###########",
                                            "#.aaa..a.a.a.a.#",
                                            "#.aaaaa..#####a#",
                                            "#a.aaaaaa..###.#",
                                            "#..aaaaaa...##a#",
                                            "#a.aaaaa.#####.#",
                                            "#.#.aaaaa.####a#",
                                            "#a#.aa.aaa.###.#",
                                            "#.##..#..aa.##a#",
                                            "#a##.####.aa.#.#",
                                            "#.########.aa.a#",
                                            "#a#########.aa..",
                                            "#.a.a.a.a.a..a.#",
                                            "#############.##" ] ), 1, 1 )

    def activate( self ):
        if len(self.canvas.layers()) > 0:
            self.activeLayer = self.canvas.layers()[0]

            if not self.activeLayer.type() == 0: # Not a vector layer
                self.rubberBand.hide()  
                QMessageBox.warning( self.canvas, "Warning", 
                    "The first layer has to be a vector layer with lines" )
                return

            self.canvas.setCursor( self.cursor )
        else:
            QMessageBox.warning( self.canvas, "Warning", 
                    "Please load a layer to the map" )

    def canvasPressEvent( self, event ):

        if self.activeLayer == None:
            return

        p = QPoint( event.pos().x(), event.pos().y() )
        tolerancia = 3 # For lines

        # Armar una caja alrededor del punto picado
        pmin = self.toMapCoordinates( QPoint( p.x() - tolerancia,
            p.y() + tolerancia ) )
        pmax = self.toMapCoordinates( QPoint(p.x() + tolerancia,
            p.y() - tolerancia ) )
        bBox = QgsRectangle( pmin.x(), pmin.y(), pmax.x(), pmax.y() )

        provider = self.activeLayer.dataProvider()
        feat = QgsFeature()

        # Obtener geometría, intersección espacial con precisión
        provider.select( [], bBox, True, True)

        featCount = 0

        # Obtiene los atributos de cada feature y los almacena en un diccionario
        while provider.nextFeature( feat ):
            featCount += 1    
            break 

        if featCount == 0:
            self.rubberBand.reset( False ) # Borrar RubberBand
        else:
            # Inicializar RubberBand
            self.rubberBand.reset( feat.geometry().type() == QGis.Polygon )
            self.rubberBand = QgsRubberBand( self.canvas, feat.geometry().type() == QGis.Polygon )

            if ( self.rubberBand ): # Definir y mostrar RubberBand
                self.rubberBand.setToGeometry( feat.geometry(), self.activeLayer )
                self.rubberBand.setWidth( 2 )
                self.rubberBand.setColor( Qt.red )
                self.rubberBand.show()

            if self.bStartingNode:
                self.emit( SIGNAL( "startingFeatureSelected" ), feat.id() )
            else:
                self.emit( SIGNAL( "targetFeatureSelected" ), feat.id() )

    def deactivate( self ):
        """ Borra el rubberBand y cierra el diálogo """
        #self.rubberBand.reset( False )
        pass

                    

